"""
encapsulate an action, and the registering of actions
"""
from __future__ import absolute_import
import logging

log = logging.getLogger(__name__)

_actions = {}

class Action(object):
    """
    enapsulates a an action
    """

    def __init__(self, name, method, priority):
        """
        initialize the action
        """
        self.name = name
        self.method = method
        self.priority = priority

    def __call__(self, *args, **kwargs):
        """
        override call to leverage the registered method
        """
        return self.method(*args, **kwargs)

    def __eq__(self, other):
        """
        override eq to compare priority
        """
        return self.priority == other.priority

    def __lt__(self, other):
        """
        override lt to compare priority
        """
        return self.priority < other.priority

    def __gt__(self, other):
        """
        override gt to compare priority
        """
        return self.priority > other.priority

    def __str__(self):
        """
        override str to provide a readable view of action
        """
        return '<Action(name=%s,method=%s,priority=%s)>' % (self.name, 
            self.method.__name__, self.priority)

    __repr__ = __str__

# end of Action

def add_action_handler(name, method, priority):
    """
    creates an action with associated method and priority
    """
    actions = _actions.setdefault(name, [])
    for action in actions:
        if action.method == method:
            raise Exception('%s has already been registered as action under name %s' % (method.__name__, name))
    log.debug('registered method %s to action %s', method.__name__, name)
    action = Action(name, method, priority)
    actions.append(action)
    return action


def get_actions(name):
    """
    @UNUSED
    retrieves a list of Action based on specified name
    """
    if not name in _actions:
        raise KeyError('No such action %s' % name)
    _actions[name].sort(reverse=True)
    return _actions[name]


def exec_action(name, *args, **kwargs):
    """
    @UNUSED
    execute the associated action based on name action
    """
    log.trace('Time to execute action named: [%s]', name)
    if not name in _actions:
        return
    for action in get_actions(name):
        log.trace('Using action: [%s]', action)
        action(*args, **kwargs)



""" Plugin Loading & Management.
"""
from __future__ import absolute_import
import os, sys
import re
import logging
import time
import pkgutil
import itertools
import ordereddict

from seedbox.action import add_action_handler
from seedbox import processmap
from seedbox import tasks as plugins_pkg

log = logging.getLogger(__name__)

__all__ = ['PluginWarning', 'PluginError', 'register_plugin', 'phase',
           'get_plugin_by_name', 'get_plugins']


class DependencyError(Exception):
    """Plugin depends on other plugin, but it cannot be loaded.

    Args:
        issued_by: name of the plugin trying to do the import
        missing: name of the plugin or library that is missing
        message: user readable error message

    All args are optional.
    """

    def __init__(self, issued_by=None, missing=None, message=None, silent=False):
        super(DependencyError, self).__init__()
        self.issued_by = issued_by
        self.missing = missing
        self._message = message
        self.silent = silent

    def _get_message(self):
        """
        retrieve message
        """
        if self._message:
            return self._message
        else:
            return 'Plugin `%s` requires dependency `%s`' % (self.issued_by, self.missing)

    def _set_message(self, message):
        """
        sets message
        """
        self._message = message

    def has_message(self):
        """
        checks message exists
        """
        return self._message is not None

    message = property(_get_message, _set_message)

    def __str__(self):
        return '<DependencyError(issued_by=%r,missing=%r,message=%r,silent=%r)>' % \
            (self.issued_by, self.missing, self.message, self.silent)


class DisabledPluginError(Exception):
    """
    enables standard error messages for plugins
    """
    def __init__(self, value, logger=log, **kwargs):
        super(DisabledPluginError, self).__init__()
        self.value = value
        self.log = logger
        self.kwargs = kwargs

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return self.value

class PluginWarning(Warning):
    """
    enables standard warning messages for plugins
    """
    def __init__(self, value, logger=log, **kwargs):
        super(PluginWarning, self).__init__()
        self.value = value
        self.log = logger
        self.kwargs = kwargs

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return self.value


class PluginError(Exception):
    """
    enables standard error messages for plugins
    """
    def __init__(self, value, logger=log, **kwargs):
        super(PluginError, self).__init__()
        self.value = value
        self.log = logger
        self.kwargs = kwargs

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return self.value

DEFAULT_PRIORITY = 128

def phase(name, priority=DEFAULT_PRIORITY):
    """
    Set the phase and priority assoication for plugin
    """
    def decorator(target):
        """
        adds phase_name and priority variable to function/method
        """
        target.phase_name = name
        target.priority = priority
        return target
    return decorator


# Mapping of plugin name to PluginInfo instance (logical singletons)
plugins = {}

# holds reference to configurations
configurator = None

class PluginInfo(dict):
    """
    Allows accessing key/value pairs of this dictionary subclass via
    attributes. Also instantiates a plugin and initializes properties.
    """
    # Counts duplicate registrations
    dupe_counter = 0

    def __init__(self, plugin_class, name=None):
        """
        Register a plugin.
        """
        dict.__init__(self)

        if name is None:
            # Convention is to take camel-case class name and rewrite it to an 
            # underscore form, e.g. 'PluginName' to 'plugin_name'
            name = re.sub('[A-Z]+', lambda i: '_' + i.group(0).lower(), plugin_class.__name__).lstrip('_')

        # Set basic info attributes
        self.name = name
        self.phase_handlers = {}

        # Create plugin instance
        self.plugin_class = plugin_class
        self.plugin_module = (plugin_class.__module__).rsplit('.', 1).pop()
        self.configs = configurator.get_configs(self.plugin_module)
        self.instance = self.plugin_class()
        self.instance.plugin_info = self  # give plugin easy access to its own info

        if self.name in plugins:
            PluginInfo.dupe_counter += 1
            log.critical('Error while registering plugin %s. %s' %
                         (self.name, ('A plugin with the name %s is already registered' % self.name)))
        else:
            # don't even load the plugin if disabled
            if self.configs.disabled:
                log.info('during configuration %s was disabled', self.name)
                raise DisabledPluginError('disabled plugin based on configuration (%s)' % self.name)

            self.map_action_methods()
            plugins[self.name] = self


    def map_action_methods(self):
        """
        Search through the provided class and determine which methods are associated
        with an action and generate an Action for the method. Ignoring all 
        private '_' methods. 
        """
        import inspect

        plugin_valid = False
        log.trace('Check plugin for action methods [%s]', self.plugin_class.__name__)
        for class_method in inspect.getmembers(self.plugin_class, predicate=inspect.ismethod):
            method_name = class_method[0]
            log.trace('method=[%s]', method_name)
            # ignore any method name that is private, doesn't matter if it is
            # single or double underscore. Just go to next method
            if method_name.startswith('_'):
                log.trace('method=[%s] is private; skipping', method_name)
                continue

            # now we need to convert method_name to an object, so that we
            # can access it and retrieve the decorated attributes; since we
            # just got the name from the class itself, there is no need to
            # check hasattr or set a default value on getattr
            method = getattr(self.instance, method_name)
            if not callable(method):
                log.trace('method=[%s] is not callable; skipping', method_name)
                continue

            # retrieve values from action decorator if they exist;
            # if they are not found, or have used an incorrect type or value
            # then we will ignore the corresponding error and just continue
            # to the next method in the list
            try:
                log.trace('method=[%s] checking for decorated attributes', method_name)
                handler_prio = _check_value('priority', method.priority)
                handler_phase = _check_value('phase', method.phase_name)
                if not handler_phase in processmap.get_run_phases():
                    raise ValueError(
                        'attribute [%s] has unsupported value [%s]; supported values: %s' % ('phase', 
                            handler_phase, processmap.get_run_phases()))
                log.trace('method=[%s] decorated attributes found', method_name)
                plugin_valid = True
            except ValueError as valerr:
                log.debug('method=[%s] decorated attribute has ValueError --> %s', method_name, valerr)
                continue
            except TypeError as tyerr:
                log.debug('method=[%s] decorated attribute has TypeErorr --> %s', method_name, tyerr)
                continue
            except AttributeError as aterr:
                log.debug('method=[%s] not decorated; AttributeError --> %s', method_name, aterr)
                continue
            except Exception as err:
                log.debug('method=[%s] not decorated; generic error --> %s', method_name, err)
                continue

            # now create the action and link phase to action
            action = add_action_handler('%s.%s' % (handler_phase, self.name), method, handler_prio)
            self.phase_handlers[handler_phase] = action
            log.trace('method=[%s] linked to action [%s] with priority [%d]', 
                method_name, action.name, handler_prio)

        if not plugin_valid:
            errmsg = ('class [%s] is not a valid plugin; no methods defined for a supported phase. '
                'Check @phase for proper configuration.' % self.plugin_class.__name__)
            log.trace(errmsg)
            raise PluginError(errmsg)


    def __getattr__(self, attr):
        """
        override getattr
        """
        if attr in self:
            return self[attr]
        return dict.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        """
        override setattr
        """
        self[attr] = value

    def __str__(self):
        """
        override str
        """
        return '<PluginInfo(name=%s)>' % self.name

    __repr__ = __str__


register_plugin = PluginInfo

def _check_value(name, val):
    """
    Need to verify if we have actual values of the correct type
    provided as inputs to our action decorator
    """
    if not isinstance(val, (basestring, int)):
        raise TypeError('attribute [%s] value not string/int: [%s]' % (name, type(val).__name__))

    if isinstance(val, basestring):
        if not val:
            raise ValueError('attribute [%s] has no value [%s]' % (name, val))

        use_val = val.strip()
        if not use_val:
            raise ValueError('attribute [%s] has empty value [%s]' % (name, use_val))
    else:
        # this means it is of type int, but just in case we will call
        # int function to make sure so we get ValueError if it is not
        # an actual int
        use_val = int(val)

    return use_val


def _strip_trailing_sep(path):
    """
    simple function to strip trailing separator
    """
    return path.rstrip("\\/")


def get_standard_plugins_path(user_paths=None):
    """
    Generates a list of locations from where to load plugins
    currently our tasks package is the only location for accessing plugins

    If the user provides specific paths, then we will attempt to access those
    and load those plugins as well. 
    args:
        user_paths: the directory paths provided by user to search for plugins
    returns:
        list of directory paths where to access plugins
    """
    # Use standard default
    paths = []

    # Add tasks directory (plugins)
    paths.append(os.path.abspath(os.path.dirname(plugins_pkg.__file__)))

    # now add any paths provided by user
    for user_path in user_paths:
        if os.path.exists(user_path):
            paths.append(user_path)
        else:
            log.warning('plugin-path specified not found [%s]', user_path)

    return paths


def _load_plugins_from_dirs(dirs):
    """
    performs the actual loading of plugins from specified directory
    """

    log.trace('Trying to load plugins from: %s' % dirs)
    # add all dirs to plugins_pkg load path so that plugins are loaded 
    # from provided directories
    plugins_pkg.__path__ = map(_strip_trailing_sep, dirs)
    for importer, name, ispkg in pkgutil.walk_packages(dirs, plugins_pkg.__name__ + '.'):
        if ispkg:
            continue
        # Don't load any plugins again if they are already loaded
        # This can happen if one plugin imports from another plugin
        if name in sys.modules:
            continue
        loader = importer.find_module(name)
        # Don't load from pyc files
        if not loader.filename.endswith('.py'):
            continue
        try:
            loaded_module = loader.load_module(name)
        except DependencyError as deperr:
            if deperr.has_message():
                msg = deperr.message
            else:
                msg = 'Plugin `%s` requires `%s` to load.' % (deperr.issued_by or name, 
                    deperr.missing or 'N/A')
            if not deperr.silent:
                log.warning(msg)
            else:
                log.debug(msg)
        except DisabledPluginError as dperr:
            log.warning('Plugin %s was not loaded. %s', name, dperr)
        except PluginError as plugerr:
            log.critical('Plugin %s was not loaded. %s', name, plugerr)
            log.exception(plugerr)
        except ImportError as imperr:
            log.critical('Plugin `%s` failed to import dependencies' % name)
            log.exception(imperr)
        except Exception as err:
            log.critical('Exception while loading plugin %s' % name)
            log.exception(err)
            raise
        else:
            log.debug('Loaded module %s from %s' % (name, loaded_module.__file__))


def load_plugins(configmgr, user_paths):
    """Load plugins from the standard plugin paths."""
    global configurator
    configurator = configmgr

    start_time = time.time()
    _load_plugins_from_dirs(get_standard_plugins_path(user_paths))
    took = time.time() - start_time
    log.trace('Plugins took %.2f seconds to load' % took)

    # generate a map by phase of each of the plugins we just
    # loaded so the Manager knows what we have to offer
    phasemap = ordereddict.OrderedDict()
    for phase in processmap.get_run_phases():
        phasemap[phase] = get_plugins(phase)

    return phasemap


def get_plugins(phase=None):
    """
    Retrieve a list of plugins; all if no input provided else those that match
    the criteria provided 

    args:
        phase: the name of the phase that a plugin supports (default: None)

    returns:
        iterable of PlugInfo instances
    """
    def matches(plugin):
        """
        filters out plugins that don't match
        """
        if phase and not phase in plugin.phase_handlers:
            log.trace('Phase resulted in filtering out plugin: [%s]', plugin)
            return False
        return True
    return itertools.ifilter(matches, plugins.itervalues())


def get_plugin_by_name(name):
    """
    Get plugin by name, preferred way since this structure may be changed at some point.
    """
    if not name in plugins:
        raise DependencyError(missing=name, message='Unknown plugin %s' % name)
    return plugins[name]




"""
Manages all Options for system.
"""
from __future__ import absolute_import
from argparse import ArgumentParser, SUPPRESS
from configparser import ConfigParser
import logging
import os, sys

from seedbox import tools

log = logging.getLogger(__name__)

DEFAULT_CFG_FILENAME = 'seedbox.cfg'
DEFAULT_LOG_FILENAME = 'seedbox.log'
DEFAULT_LOG_LEVEL = 'info'

class _Namedspace(object):
    """
    simple namespace holder
    """
    def __init__(self, **kwargs):
        """
        This is how we initialize our instance variables. If we are provided with
        any inputs then we will add those as attributes to ourself.

        Side-effect: A flag was needed to identify if we have been configured or
            not because we are leveraged by argparse to set any command line
            arguments onto us. So hasattr has to get an exception when an
            attribute does not exist. But in general use after we have been
            configured, we just want to simply return None when the attribute
            is missing. Therefore we set our hidden flag to false initially, and
            once we are configured, we will flip the flag to true. For those
            instances that pass in values to start with, we will go ahead and flip
            the flag for them. For those that manually set attributes via 
            setattr, they will need to manually flip the flag via configured()
            method.
        """
        self.__configured = False
        if kwargs:
            self.add(**kwargs)

    __hash__ = None

    def add(self, **kwargs):
        """
        Provides the ability to add a list of attributes to ourself in 
        bulk other than during init(). Even init() will call us so that
        not matter when this happens it is consistent.
        """
        self.__configured = True
        for name in kwargs:
            setattr(self, name, kwargs[name])

    def configured(self):
        """
        handles the manual flipping of the flag after configuration is completed.
        """
        self.__configured = True

    def __getattr__(self, name):
        """
        The normal functionality for this would be to throw an
        AttributeError if the name does not exist in the dict.
        But we are scoping this to represent information from
        within configuration file. That means some sections might
        not have the same set of attributes. Instead of each piece
        of code trying to check for AttributeError, we will perform
        a get on the dict and therefore return None when the name/key
        is not within our attribute dict.
        Update: Need to raise the exception as normal otherwise hasattr
        will cause issues as argparse adds 'hidden' attributes in some
        scenarios. And then tries to remove them. So if hidden, then we
        will raise exception.
        """
        if self.__configured:
            return self.__dict__.get(name)
        return object.__getattribute__(self, name)

    def __eq__(self, other):
        """
        provide implementation
        """
        return vars(self) == vars(other)

    def __ne__(self, other):
        """
        provide implementation
        """
        return not (self == other)

    def __contains__(self, key):
        """
        provide implementation
        """
        return key in self.__dict__

    def __repr__(self):
        """
        provide implementation
        """
        type_name = type(self).__name__
        arg_strings = []
        for arg in self._get_args():
            arg_strings.append(repr(arg))
        for name, value in self._get_kwargs():
            # if I went through the trouble to 'hide' it by using
            # double underscore, I do not want it as part of the visible
            # representation of the instance either!
            if name.startswith('_Namedspace__'):
                continue
            arg_strings.append('%s=%r' % (name, value))
        return '%s(%s)' % (type_name, ', '.join(arg_strings))

    def _get_kwargs(self):
        """
        provide implementation
        """
        return sorted(self.__dict__.items())

    def _get_args(self):
        """
        provide implementation
        """
        return []

# end _Namedspace

class ConfigOptions(_Namedspace):
    """
    Create a combined solution for brining both arparse and configparser together.
    I prefer being able to access the values like an attribute vs. figure out
    what section I need, then check if key exists, and then convert to correct type.

    Also it would be nice to simply pass a method a set based on namespace (module name)
    vs. passing the entire configuration set which is not applicable for most areas in the
    code.
    """

    def __init__(self, appversion):
        """
        process command line inputs, using those inputs and some defaults; attempt
        to load config file. That is where the magic will start, then we will need
        to merge the results together and generate a namespace for each of the 
        'sections' identified within configuration file.
        """
        super(ConfigOptions, self).__init__()
        self.appversion = appversion
        self._namedspaces = {}
        self._basespace = _Namedspace()
        self._configparser = ConfigParser(allow_no_value=True)

        _load_args(self._basespace, appversion)
        # need to set the flag that we are configured so accessing attributes
        # results in a value. See parent class for details.
        self._basespace.configured()

    def _generate_namespaces(self):
        """
        generate namespaces for configs
        """
        # to use interpolation, we have to call items on the parser otherwise we get back
        # just raw data with all the placeholders showing up
        self._basespace.add(**_validate_and_format(self._configparser.items('DEFAULT')))

        for sect in self._configparser.sections():
            self._namedspaces[sect] = _Namedspace(**_validate_and_format(self._configparser.items(sect)))

    def load_configs(self):
        """
        Now we are ready to load the configuration files
        """
        _find_configs(self._configparser, self._basespace)
        self._generate_namespaces()

    def get_configs(self, module_name=None):
        """
        return configs as requested
        """
        return self._namedspaces.get(module_name, self._basespace)


# end ConfigOptions

def _set_type(value):
    """
    a workaround for a problem where we set flags using strings in a config file
    but we really need them to be booleans when we consume them. Also handles
    converting a value to None if no value; otherwise a string.
    """
    if value.lower() in tools.BOOLEAN_STATES:
        return tools.BOOLEAN_STATES[value.lower()]
    else:
        if not value:
            return None
        return str(value)

def _verify_paths(path_list):
    """
    verify list of paths; return list -- it will also handle splitting any string list
    that is provided vs. an actual list. 
    """
    to_verify_paths = tools.to_list(path_list)

    valid_items = []
    for item in to_verify_paths:
        verified_item = tools.verify_path(item)
        if verified_item:
            valid_items.append(verified_item)

    return valid_items

def _validate_and_format(configs):
    """
    convert the items within each section of the configuration file into a dict
    while also converting the value to a proper type (boolean, list, str, etc.)
    Also validate that required common entries are present, and convert knowns
    paths into absolute paths when they are relative, while also removing any
    non-existent paths to avoid issues later.

    args:
        configs: the list of tuples (items) from each section of the configuration
            file that needs to be converted to dict.
    exception:
        ValueError: if any of the required entries within the configuration
            file are not found or found to be missing or invalid.
    """
    config_group = {}
    missing_required = []

    for key, value in configs:
        config_group[str(key)] = _set_type(value)
    log.trace('updated group: %s', config_group)


    # Need to verify the path and convert to absolute path if relative.
    # if the entry is missing all together, then None will be sent for
    # verification and None will be return. So the entry will be always
    # a full and complete path or None. If None then we'll raise an error
    # down below. 
    config_group['torrent_path'] = tools.verify_path(config_group.get('torrent_path'))
    if not config_group.get('torrent_path'):
        log.debug('torrent_path was not supplied in the configuration file')
        missing_required.append('TORRENT_PATH')

    config_group['incomplete_path'] = tools.verify_path(config_group.get('incomplete_path'))
    if not config_group.get('incomplete_path'):
        log.debug('incomplete_path was not supplied in the configuration file')
        missing_required.append('INCOMPLETE_PATH')

    config_group['media_paths'] = _verify_paths(config_group.get('media_paths'))
    if not config_group.get('media_paths'):
        log.debug('media_paths was not supplied in the configuration file')
        missing_required.append('MEDIA_PATHS')

    # plugin paths is optional; so if it exists it will be a list separated
    # by commas. Unable to split None, so the default will be an empty string
    # which when split will result in an empty list. If the verified list is
    # empty then the verified list will also be empty. If the items in the list
    # are not valid then it will again be an empty list.
    config_group['plugin_paths'] = _verify_paths(config_group.get('plugin_paths', ''))

    # disabled phases is optional; so if it exists it will be a list separated
    # by commas. Unable to split None, so the default will be an empty string
    # which when split will result in an empty list. If the verified list is
    # empty then the verified list will also be empty.
    config_group['disabled_phases'] = tools.to_list(config_group.get('disabled_phases', ''))

    log.trace('final config group: %s', config_group)
    if missing_required:
        raise ValueError('missing required input(s) [%s] not supplied in configuration file)' % missing_required)

    return config_group


def _find_configs(parser, args):
    """
    look for the config file in all the common locations so we can load it
    and figure out how we should execute

    args:
        parser: configparser instance that will be used to process the actual 
            configuration file once we find it of course.
        args: any inputs we got back from the command line that will used to find
            the configuration file (rcfile or resource_path)
    exceptions:
        IOError: if no configuration file found
    """

    log.trace('ready to load configuration files')

    possible = []
    # if the user specified a path within rcfile then we will try to use it;
    # if there is no dirname that means it is just the default filename only.
    if os.path.dirname(args.rcfile):
        log.debug('using command line specified configuration file %s', args.rcfile)
        if os.path.isabs(args.rcfile):
            # explicit path given, don't try anything too fancy
            possible.append(args.rcfile)
        else:
            # specified sure, but it was a relative path so we need 
            # expand it out otherwise things will fail
            possible.append(os.path.realpath(args.rcfile))
    elif args.resource_path and os.path.exists(args.resource_path):
        log.debug('using command line specified resource location %s', args.resource_path)
        # so we will look in this location only
        if os.path.isabs(args.resource_path):
            possible.append(args.resource_path)
        else:
            # specified sure, but it was a relative path so we need 
            # expand it out otherwise things will fail later
            possible.append(os.path.realpath(args.resource_path))
    else:
        log.debug('no configuration details specified, checking usual locations')
        # normal lookup locations
        possible.append(os.path.dirname(os.path.abspath(sys.path[0])))
        possible.append(os.getcwd())
        possible.append(sys.path[0])
        possible.append(os.path.join(os.path.expanduser('~'), '.seedbox'))
        if sys.platform.startswith('win'):
            # On windows look in ~/seedbox as well, as explorer does not let you 
            # create a folder starting with a dot
            possible.append(os.path.join(os.path.expanduser('~'), 'seedbox'))

    log.trace('possibles: %s', possible)
    for location in possible:
        log.debug('checking location %s', location)
        # if the user specified an actual location and file via rcfile
        # then there is nothing to add, but if we are looking at a location
        # only (ie directory) then we need to append the default filename
        if os.path.isdir(location):
            config_file = os.path.join(location, DEFAULT_CFG_FILENAME)
        else:
            config_file = location
        log.trace('checking file %s', config_file)
        if os.path.exists(config_file):
            log.trace('file exists: %s', config_file)
            # load it up
            found = parser.read(config_file)
            if not found:
                log.warn('loading of config file %s had no results; both rare and strange', config_file)
                continue
            log.debug('loaded config file found: [%s]', found)
            # need to set the resource path (include in args and configuration file)
            parser.set('DEFAULT', 'resource_path', os.path.dirname(config_file))
            args.resource_path = os.path.dirname(config_file)
            log.debug('setting resource path to %s', os.path.dirname(config_file))
            # if this break is never reached then the else block will be executed
            # and the corresponding exception will be raised
            break
    else:
        log.error('tried to read from %s but no configuration file found', possible)
        raise IOError('no configuration file found; searched %s' % possible)

def _load_args(namespace, app_version):
    """
    setups what command arguments there are, and then processes the input on execution

    args:
        namespace: a class instance that will hold all the command arguments as attributes
            once the command has been processed.
        app_version: the version of the application
    returns:
        N/A
    """

    prog_desc = 'This program is the main interface for executing a series of \
        tasks in sequence. The types of task are intended for managing \
        torent files and the associated media files on a seedbox.'

    parser = ArgumentParser(description=prog_desc)

    parser.add_argument('--version', action='version',
        version='%(prog)s '+app_version)

    parser.add_argument('--resource', dest='resource_path', metavar='RESOURCE_PATH',
        help='path where all resource files are located; config file, logfile, etc.')
    parser.add_argument('--rcfile', dest='rcfile', metavar='RESOURCE_FILE', default=DEFAULT_CFG_FILENAME,
        help='specify a configuration file (location is resource path)')

    parser.add_argument('--logfile', dest='logfile', metavar='LOG_FILE', default=DEFAULT_LOG_FILENAME,
        help='specify name of log file (location is resource path)')
    parser.add_argument('--loglevel', dest='loglevel', metavar='LOG_LEVEL', default=DEFAULT_LOG_LEVEL,
        choices=['none', 'critical', 'error', 'warning', 'info', 'debug', 'trace'],
        help=SUPPRESS)

    parser.add_argument('--reset', action='store_true',
        help='DANGER: deletes the database cache and everything starts over')
    parser.add_argument('--retry', action='store_true',
        help='only executes entries that failed previously')

    parser.add_argument('--dev', action='store_true', help=SUPPRESS)

    log.trace('all known arguments added to parser....now parse')
    # the namespace is an object that will actually hold a reference to each value in
    # a corresponding attribute field. dest=attribute name; we are parsing the 
    # known args ONLY; we will even ignore those that were extraneous.
    parser.parse_known_args(namespace=namespace)
    log.debug('found command line arguments: [%s]', namespace)


def initialize(appversion):
    """
    just a convience method, and allows us to expand over time as required w/o changing
    our manager module.
    creates an instance of ConfigOptions which will be our handle to arguments passed
    on the command line, and later references to the namespaces holding the
    configuration file that is loaded.

    args:
        appversion: string representing the current version of application. Used as part of help
            and --version commands.
    returns:
        and instance ConfigOptions which will act as the handle/container for all
        options/configurations provided via command line and configuration file.
    """
    log.trace('initializing setup...processing command line arguments')
    return ConfigOptions(appversion)




"""
Handles mapping
"""
from __future__ import absolute_import
import logging
import ordereddict
from collections import namedtuple
from seedbox.workflow import Taskflow

log = logging.getLogger(__name__)

# leverage namedtuple to hold ordered list with named attributes for
# simple and effective access while using lease amount of memory
_TaskInfo = namedtuple('TaskInfo', ['task', 'action'])
_Phase = namedtuple('Phase', ['name', 'tasks'])

# initialize the map to an empty ordered dictionary
# order will be important
_procmap = ordereddict.OrderedDict()

# using our defined Workflow and the related transistions/phases
# generate a list of phases (in order: see tuple) 
# based on existing configuration we will remove the last item which
# is the cancel phase which is not runnable ;-)
_run_phases = tuple([tran.name for tran in Taskflow.transitions][:-1])

# could use a function, but to avoid creating new variables over and over
# again just to access a name value of a state for a specified phase seemed
# overkill, so using a lambda. Provides a simple mapping of phase to state.
# based on configuration we know it is always only one related state
_phase_to_state = lambda phase: (Taskflow.transitions[phase]).source[0].name

_state_to_phase = lambda state: list(Taskflow.transitions.available_from(state))[0].name


def get_tasks(state):
    """
    retrieve the task details
    """
    return _procmap[state]


def get_state(phase):
    """
    determine my state
    """
    return _phase_to_state(phase)

def get_phase(state):
    """
    determine my phase
    """
    return _state_to_phase(state)

def get_run_phases():
    """
    Retrieves a list of run phases (ie those phases where execution can/will occur)

    args:
        N/A
    returns:
        an ordered tuple of phase names
    """
    return _run_phases

def init(phasemap):
    """
    initializes the map
    """

    for phase in phasemap.keys():                
        task_list = []
        for task in sorted(phasemap[phase], key=lambda plug: plug.phase_handlers[phase], reverse=True):
            task_list.append(_TaskInfo(task, task.phase_handlers[phase]))
        
        _procmap[_phase_to_state(phase)] = _Phase(phase, tuple(task_list))

    log.trace('proc_map = %s', _procmap)



"""
Provide a simply extension to lockfile to support a Timeout for PIDFileLock
"""
from __future__ import absolute_import
from lockfile.pidlockfile import PIDLockFile

class TimeoutPIDLockFile(PIDLockFile, object):
    """ Lockfile with default timeout, implemented as a Unix PID file.

        This uses the ``PIDLockFile`` implementation, with the
        following changes:

        * The `acquire_timeout` parameter to the initialiser will be
          used as the default `timeout` parameter for the `acquire`
          method.

        """

    def __init__(self, path, acquire_timeout=None, *args, **kwargs):
        """ Set up the parameters of a TimeoutPIDLockFile. """
        self.acquire_timeout = acquire_timeout
        super(TimeoutPIDLockFile, self).__init__(path, *args, **kwargs)

    def acquire(self, timeout=None, *args, **kwargs):
        """ Acquire the lock. """
        if timeout is None:
            timeout = self.acquire_timeout
        super(TimeoutPIDLockFile, self).acquire(timeout, *args, **kwargs)




"""
A class that defines a torrent file and the associated state. Also defined is a class
that defines the contents of a torrent file, ie some media files typically.
"""
from __future__ import absolute_import
#from sqlobject import *
import sqlobject
import logging
import os
import shutil

log = logging.getLogger(__name__)
DB_NAME = 'torrent.db'
MAX_BACKUP_COUNT = 12

class Torrent(sqlobject.SQLObject):
    """
    A class that defines a torrent and associated database table using
    SQLObjects to handle persistence.
    """
    name = sqlobject.StringCol(unique=True)
    create_date = sqlobject.DateCol(default=sqlobject.DateTimeCol.now)
    state = sqlobject.EnumCol(enumValues=['init', 'ready', 'active', 'done', 'cancelled'], default='init')
    retry_count = sqlobject.IntCol(default=0)
    failed = sqlobject.BoolCol(default=False)
    error_msg = sqlobject.StringCol(default=None)
    invalid = sqlobject.BoolCol(default=False)
    purged = sqlobject.BoolCol(default=False)
    media_files = sqlobject.SQLMultipleJoin('MediaFile')

class MediaFile(sqlobject.SQLObject):
    """
    A class that defines a media file contained within a torrent file
    and associatd database table using SQLObjects to handle persistence.
    """

    filename = sqlobject.StringCol()
    file_ext = sqlobject.StringCol()
    file_path = sqlobject.StringCol(default=None)
    size = sqlobject.IntCol(default=0)
    compressed = sqlobject.BoolCol(default=False)
    synced = sqlobject.BoolCol(default=False)
    missing = sqlobject.BoolCol(default=False)
    skipped = sqlobject.BoolCol(default=False)
    torrent = sqlobject.ForeignKey('Torrent', cascade=True)

class AppState(sqlobject.SQLObject):
    """
    A class that defines maintains information about the state of the application
    and internal processing. Effectively it is just a key-value pair except we
    store the values by type.
    """

    name = sqlobject.StringCol(unique=True)
    val_str = sqlobject.StringCol(default=None)
    val_int = sqlobject.IntCol(default=-99999)
    val_list = sqlobject.StringCol(default=None)
    val_flag = sqlobject.BoolCol(default=False)
    val_date = sqlobject.DateTimeCol(default=sqlobject.DateTimeCol.now)


def init(resource_path, reset=False):
    """
    Establish a connection and make sure all structures are in place
    """

    dbloc = os.path.join(resource_path, DB_NAME)
    log.trace('location of database will be: [%s]', dbloc)
    db_exists = False
    if os.path.exists(dbloc):
        db_exists = True
        log.info('loading database [%s]', dbloc)
    else:
        log.info('database does not exist; creating database....[%s]', dbloc)

    connect_str = '{0}{1}{2}'.format('sqlite:', dbloc, '?driver=sqlite3')
    log.trace('establishing connection to database: [%s]', connect_str)
    # create connection to the database
    sqlobject.sqlhub.processConnection = sqlobject.connectionForURI(connect_str)

    # if the database already existed and a reset was requested then
    # we drop all the tables
    if db_exists and reset:
        log.info('reset requested; deleting database cache.')
        Torrent.dropTable(ifExists=True)
        MediaFile.dropTable(ifExists=True)
        AppState.dropTable(ifExists=True)
        log.info('recreating database cache....')

    # we have defined the class, so create the table if it doesn't exist
    Torrent.createTable(ifNotExists=True)

    # we have defined the class, so create the table if it doesn't exist
    MediaFile.createTable(ifNotExists=True)

    # we have defined the class, so create the table if it doesn't exist
    AppState.createTable(ifNotExists=True)

    if log.isEnabledFor(logging.DEBUG):
        log.trace('Torrent table: [%s]', Torrent.sqlmeta.table)
        log.trace('Torrent Columns: [%s]', Torrent.sqlmeta.columns.keys())
    
        log.trace('MediaFile table: [%s]', MediaFile.sqlmeta.table)
        log.trace('MediaFile Columns: [%s]', MediaFile.sqlmeta.columns.keys())

        log.trace('AppState table: [%s]', AppState.sqlmeta.table)
        log.trace('AppState Columns: [%s]', AppState.sqlmeta.columns.keys())


def backup(resource_path):
    """
    create a copy of myself
    """
    log.trace('starting database backup process')
    default_db_name = os.path.abspath(os.path.join(resource_path, DB_NAME))
    log.trace('location of database: [%s]', default_db_name)
    if os.path.exists(default_db_name):
        for i in range(MAX_BACKUP_COUNT - 1, 0, -1):
            sfn = '%s.%d' % (default_db_name, i)
            dfn = '%s.%d' % (default_db_name, i + 1)
            if os.path.exists(sfn):
                if os.path.exists(dfn):
                    os.remove(dfn)
                os.rename(sfn, dfn)
        dfn = default_db_name + '.1'
        log.info('backing up db [%s] to [%s]', default_db_name, dfn)
        shutil.copy2(default_db_name, dfn)
        log.info('backup complete')
    else:
        log.warn('Database [%s] does not exist, no backup taken.', default_db_name)




"""
workflow
"""
from __future__ import absolute_import
import xworkflows

class Taskflow(xworkflows.Workflow):
    """
    Define the workflow conditions for managing torrents;
    """
    states = (
        ('init', (u"Initial state")),
        ('ready', (u"Ready")),
        ('active', (u"Active")),
        ('done', (u"Done")),
        ('cancelled', (u"Cancelled")),
    )

    transitions = (
        ('prepare', 'init', 'ready'),
        ('activate', 'ready', 'active'),
        ('complete', 'active', 'done'),
        ('cancel', ('ready', 'active'), 'cancelled'),
    )

    initial_state = 'init'



"""
Manages the overflow of execution
"""
from __future__ import absolute_import
import logging
import xworkflows

from seedbox import processmap, datamanager
from seedbox.workflow import Taskflow

log = logging.getLogger(__name__)

DEFAULT_MAX_RETRY = 5

class ProcessError(Exception):
    """
    Identify when the Process errors
    """
    def __init__(self, value):
        super(ProcessError, self).__init__()
        self.value = value

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return self.value

class Process(xworkflows.WorkflowEnabled):
    """
    Handles workflow transistions and associated updates
    """

    state = Taskflow()

    def __init__(self):
        """
        initialize Process for a specific torrent
        """
        super(Process, self).__init__()

    @xworkflows.transition()
    def prepare(self, torrents):
        """
        handle steps necessary to make prepare transistion
        """
        return self._run(torrents)


    @xworkflows.transition()
    def activate(self, torrents):
        """
        handle steps necessary to make activate transistion
        """
        return self._run(torrents)


    @xworkflows.transition()
    def complete(self, torrents):
        """
        handle steps necessary to make complete transistion
        """
        return self._run(torrents)


    def abort(self, torrents):
        """
        handle steps necessary to make cancel transistion;
        typically called from retry logic
        """
        log.trace('current state [%s]', self.state.name)
        self.cancel()
        log.trace('updated state [%s]', self.state.name)
        datamanager.update_state(torrents, self.state.name)


    def get_transistion_method(self):
        """
        figure out which method needs to be executed next based on current
        state to trigger a transistion in the workflow.
        """

        # we only need first one since it is a 1:1 mapping of state
        # to transition; therefore no need to go beyond first entry
        phase = processmap.get_phase(self.state)
        log.trace('processing phase [%s] for current state %s', phase, self.state.name)

        method_name = self._get_next_step_func(phase)
        log.trace('phase method name %s', method_name)
        tran_method = getattr(self, method_name)
        log.trace('phase method %s', tran_method)

        # make sure it is really callable
        if not callable(tran_method):
            log.error('method=[%s] is not callable; stopping workflow', method_name)
            raise ProcessError('no steps to execute %s of process' % phase)

        return tran_method


    def _get_next_step_func(self, tran_name):
        """
        get the function we need to execute to perform transition
        if you do not name your transition methods after the name of the transitions
        then you need to find out which function maps to which transition.
        all of this information is stored within the workflow library xworkflows
        based on what we defined previously. I'd rather not depend on the internals
        especially names of things but it is what it is and I like automation!
        """
        imp_dict = self._xworkflows_implems.get('state')

        return imp_dict.transitions_at.get(tran_name)


    def _run(self, torrents):
        """
        run the actual task for phase based on state
        """
        processed_torrents = []
        # we will only reach this point if called for purposes of advancing to
        # next state so simply return empty.
        if not torrents:
            log.info('Advancing state [%s]', self.state.name)
            # torrents is already empty but since we expect a list back
            # we will return a true empty list since torrents could have
            # been None or something else
            return processed_torrents

        phase_tasks = processmap.get_tasks(self.state.name)
        log.debug('state:phase [%s]:[%s]', self.state.name, phase_tasks.name)

        # we are going to assume that all torrents get processed successfully
        processed_torrents = torrents
        for taskinfo in phase_tasks.tasks:

            log.trace('executing action [%s] for task [%s]', taskinfo.action, taskinfo.task)
            # by passing in the processed_torrents for to a task (initially all of them)
            # we will overwrite the list with a new list of those that did get processed
            # and pass that list to the next task (if any). This means if a torrent
            # fails one task we should stop processing it therefore it will be removed
            # from the list and only those that need further processing will continue.
            processed_torrents = taskinfo.action(processed_torrents, taskinfo.task.configs)

        # we are only going to return the ones that were succesfully
        # processed such we don't move the state on those that are not
        # ready to move on.
        return processed_torrents


# end Process

def start(core_configs):
    """
    entry point for module to handle managing the end2end processing and tasking
    """
    if core_configs.retry:
        log.info('running in retry mode....no new work will be done')
        # only process those that are in an error state, and have not
        # already gone past our threshold
        _execute(core_configs.disabled_phases, True, core_configs.max_retry)
    else:
        log.debug('running normal mode')
        # time to start processing; first process all new stuff
        _execute(core_configs.disabled_phases)
        log.debug('normal mode completed, now executing retry mode')
        # then process all failed stuff by performing retry
        _execute(core_configs.disabled_phases, True, core_configs.max_retry)


def _execute(disabled_phases, retry=False, max_retry=None):
    """
    the entry point into the taskmanager
    """

    log.trace('disabled phases: %s', disabled_phases)

    for phase in processmap.get_run_phases():

        log.trace('starting execute for phase %s', phase)

        # if there was a request to disable a phase, we simply log some info
        # and continue to the next phase
        if phase in disabled_phases:
            log.info('disabled phase based on configuration (%s)', phase)
            continue

        # convert phase to state, and load corresponding torrents
        # for the specified state. the retry flag will be either True/False
        # if True, load only failed torrents
        # if False, load all non-failed torrents
        state = processmap.get_state(phase)
        log.trace('Loading torrents for state [%s]', state)
        torrents = datamanager.get_torrents_by_state(state, retry)

        if retry:
            log.trace('checking if any torrents need to be cancelled')
            # handle any and all cancellations; we will get back
            # only those torrents that still need processing
            # if nothing, it will be an empty list
            torrents = _handle_torrent_cancellation(torrents, max_retry)

        # if there are no torrents in this particular state then simply proceed
        # to the next phase
        if not torrents:
            log.info('No torrents found for state [%s]', state)
            continue

        try:
            # now execute the flow for this given state
            processed_torrents = _handle_phase_execution(state, torrents)

            if retry:
                log.trace('resetting counters for the successfully executed retried torrents')
                # the processed torrents were the succesful ones so 
                # we need to reset the retry counter
                for torrent in processed_torrents:
                    torrent.retry_counter = 0

        except ProcessError as procerr:
            log.error('Process failed! [%s]', procerr)
            break


def _handle_phase_execution(state, torrents):
    """
    Handle the actual execution flow for a given state
    """

    processed_torrents = []
    # establish a process
    process = Process()

    # as long as state != current state of process; we need to perform advancement
    while state != process.state.name:

        log.trace('process step needs to be advanced from %s to %s', process.state.name, state)

        if process.state.is_done or process.state.is_cancelled:
            log.debug('process step is either done or cancelled which means there is no more execution...quit')
            break

        tran_method = process.get_transistion_method()

        # execute the transistion method with no torrents
        # to force the transistion
        tran_method([])

    else:
        tran_method = process.get_transistion_method()

        log.trace('Processing torrents [%d] for state [%s]', len(torrents), state)
        processed_torrents = tran_method(torrents)
        log.trace('Successfully processed torrents [%d] for state [%s]', len(processed_torrents), state)
        datamanager.update_state(processed_torrents, process.state.name)

    # now return the successfully processed torrents so additional processing can 
    # take place if needed.
    return processed_torrents

    
def _handle_torrent_cancellation(torrents, max_retry):
    """
    check for torrents that have reached their max execution threshold and 
    perform the abort process on them so we can retire them from future execution
    """
    # we will allow users to configure this as per their preferences
    # but the value must be an integer greater than 0; if nothing is
    # provided by user then we default it, or if they provided some stupid
    # value then we override back to default.
    use_max_retry = DEFAULT_MAX_RETRY
    if max_retry is not None and isinstance(max_retry, int) and max_retry > 0:
        log.info('using configured max retry [%d]', max_retry)
        use_max_retry = max_retry

    log.trace('sorting torrents to abort or continue execution')
    execute_torrents = []
    abort_torrents = []
    for torrent in torrents:
        # has the the torrent reached its max retry limit?
        if torrent.retry_count >= use_max_retry:
            abort_torrents.append(torrent)
            continue

        # not reached limit then we will reset the failed flags
        # so that if it errors again we will have a new error message
        datamanager.reset_failed(torrent)

        # now increment retry counter as we are about to execute!
        torrent.retry_count += 1

        # now add it to the list of those we still need to perform
        # normal processing on.
        execute_torrents.append(torrent)


    # if there were any that we needed to abort this is when we do it
    if abort_torrents:
        log.debug('aborting torrents [%d] that reached their execution limit', len(abort_torrents))
        process = Process()
        try:
            process.abort(abort_torrents)
        except ProcessError as procerr:
            log.error('Abort process failed! [%s]', procerr)

    # now send back those that still need processing (could be just an empty list)
    return execute_torrents




"""
Holds a group of tools useful by all modules.
"""
import os
import shutil
# only need prior to python 3.3; after upgrade shutilwhich will do nothing
import shutilwhich

DEFAULT_INT = -99999

BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                  '0': False, 'no': False, 'false': False, 'off': False}

def to_bool(value):
    """
    converts the variety of possible values used to represent true/false to
    and actual recognized python boolean
    """
    if isinstance(value, bool):
        return value
    if isinstance(value, basestring) and value.lower() in BOOLEAN_STATES:
        return BOOLEAN_STATES[value.lower()]
    if isinstance(value, int) and value in [0, 1]:
        return (True if value else False)

    # if value wasn't a boolean or a boolean like value
    # we are simply going to check if not None which will result
    # in a True/False. 
    return value is not None

def to_list(value, separator=None):
    """
    converts a value (delimited most likely) to an actual list
    """
    delimiters = [',', ';']
    if separator:
        delimiters.append(separator)

    results = []
    # if we got an actual list just send it back as-is
    if isinstance(value, list):
        return value
    # if we got a string then we will need to split it
    if isinstance(value, basestring):
        # check for the following list of separators
        # if found in string, then split it; else try
        # the next one.
        for sep in delimiters:
            if value.find(sep) != -1:
                results = value.split(sep)
                break   # we got our sep now get out
        # if no results after looping try splitting based on just whitespace
        if not results:
            results = value.split()

        # strip away excess whitespace after the splits
        results[:] = map(str.strip, results)

    # we'll either send back an empty list or list based on splitting
    return results

def list_to_str(values):
    """
    converts a list to a delimited string list
    """
    result = None

    # if we get an actual list simply join and store as delimited list
    if values and isinstance(values, list):
        result = ','.join(values)

    # if the values are already a delimited list of values then use as-is
    elif values and isinstance(values, basestring) and values.find(',') != -1:
        result = values

    # if the values missing or not of correct format; we could throw a ValueError
    # but for now we will simply set the value to None
    return result

def to_int(value):
    """
    makes sure the value is an int, if not result will be a default int
    """
    if value is not None and isinstance(value, int):
        return value
    # ok so we got a string that needs to become an int
    elif isinstance(value, basestring):
        # remove any whitespace
        use_value = value.strip()
        # if we have a value then we will convert to int
        if use_value:
            try:
                use_value = int(use_value)
            except ValueError:
                #oops not really an int so setting to default
                use_value = DEFAULT_INT
        else:
            # value was not a value at all so setting to default
            use_value = DEFAULT_INT

    else:
        # not sure what was passed to us, but we are going
        # to set to default
        use_value = DEFAULT_INT

    # now just give them a result
    return use_value

def verify_path(path_entry):
    """
    verify a path, if it exists make sure it is 
    an absolute path and return. else None
    """
    if path_entry and os.path.exists(path_entry):
        if os.path.isabs(path_entry):
            return path_entry
        else:
            return os.path.realpath(path_entry)
    else:
        return None

def get_exec_path(program):
    """
    performs a which on the program to get full path for
    the specified program
    """
    return shutil.which(program)



"""
extension to logging
"""
from __future__ import absolute_import
import logging, logging.config, logging.handlers
import os, sys

# adding logging extension here given how minimal this is
# didn't seem to make sense to create a whole module for something
# so simple; if things change later we can move it then
class TraceLogger(logging.Logger):
    """
    Need to keep some low level logging messages included in case a bug
    ever pops up in the future, but this gives me the ability to separate
    tracing the propram from the actual basic debugging.
    """

    # a bit more detail than normal debug
    TRACE = 5

    def trace(self, msg, *args, **kwargs):
        """
        Log at TRACE level (more detailed than DEBUG).
        """
        self.log(TraceLogger.TRACE, msg, *args, **kwargs)


# Set our custom logger class as default; need to make sure
# the class has been defined before we set this; then add corresponding 
# level to be supported; then do the default configuring from our default
# configuration file. all done
logging.setLoggerClass(TraceLogger)
logging.addLevelName(TraceLogger.TRACE, 'TRACE')

# a simple hack to make sure as a developer you can see all trace action
# that happens prior to logging getting configured.
if '--dev' in sys.argv:
    rootlog = logging.getLogger()
    rootlog.setLevel(TraceLogger.TRACE)
    streamhandle = logging.StreamHandler(sys.stdout)
    simpleformatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    streamhandle.setFormatter(simpleformatter)
    rootlog.addHandler(streamhandle)



def configure(loglevel=logging.INFO, logfile='seedbox.log', resourceloc=None, devmode=False):
    """
    configure all logging for this execution
    """

    use_resource = resourceloc
    if not use_resource:
        if sys.platform.startswith('win'):
            # On windows look in ~/seedbox as well, as explorer does not let you create a 
            # folder starting with a dot
            default_path = os.path.join(os.path.expanduser('~'), 'seedbox')
        else:
            default_path = os.path.join(os.path.expanduser('~'), '.seedbox')

        # if the user folder exists; then we'll use it if we were not provided one
        if os.path.exists(default_path):
            use_resource = default_path
        else:
            # if all else fails then we just use the current working directory
            use_resource = os.getcwd()

    if devmode:
        logging.config.fileConfig(os.path.join(use_resource, 'logging.cfg'), disable_existing_loggers=False)
    else:
        logger = logging.getLogger()
        logger.setLevel(loglevel)

        logloc_filename = os.path.join(use_resource, logfile)
        handler = logging.handlers.RotatingFileHandler(logloc_filename, maxBytes=1000 * 1024, backupCount=9)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        handler.setFormatter(formatter)
        logger.addHandler(handler)

        # shut off logging from 3rd party frameworks
        xlogger = logging.getLogger('xworkflows')
        xlogger.setLevel(logging.ERROR)



#!/usr/bin/python
"""
Parses a torrent file and provides method to access the following attributes.
    . Tracker URL
    . Creation date
    . Client name, if any
    . For each file
        . name
        . length
        . checksum

Created on 2012-03-07

@author: mohanr
"""
from __future__ import absolute_import
from datetime import datetime
from StringIO import StringIO
import os
import string
import types


def humanize_bytes(in_bytes, precision=1):
    """Return a humanized string representation of a number of bytes.

    >>> humanize_bytes(1)
    '1 byte'
    >>> humanize_bytes(1024)
    '1.0 kB'
    >>> humanize_bytes(1024*123)
    '123.0 kB'
    >>> humanize_bytes(1024*12342)
    '12.1 MB'
    >>> humanize_bytes(1024*12342,2)
    '12.05 MB'
    >>> humanize_bytes(1024*1234,2)
    '1.21 MB'
    >>> humanize_bytes(1024*1234*1111,2)
    '1.31 GB'
    >>> humanize_bytes(1024*1234*1111,1)
    '1.3 GB'
    """
    abbrevs = (
        (1<<50L, 'PB'),
        (1<<40L, 'TB'),
        (1<<30L, 'GB'),
        (1<<20L, 'MB'),
        (1<<10L, 'kB'),
        (1, 'bytes')
    )

    if in_bytes == 1:
        return '1 byte'

    factor = None
    suffix = None
    for factor, suffix in abbrevs:
        if in_bytes >= factor:
            break

    return '%.*f %s' % (precision, in_bytes / factor, suffix)

class ParsingError(Exception):
    """ Error class representing errors that occur while parsing
    the torrent content.
    """
    def __init__(self, error_msg):
        Exception.__init__(self)
        self.error_msg = error_msg        
    
    def __str__(self):
        return repr(self.error_msg) 

class MalformedTorrentError(Exception):
    """ Error class representing a malformed torrent when required
    attributes are missing
    """
    def __init__(self, error_msg):
        Exception.__init__(self)
        self.error_msg = error_msg        

    def __str__(self):
        return repr(self.error_msg) 


class TorrentParser(object):
    """
    Parses a torrent file and returns various properties based on the 
    content of the torrent file.
    """
        
    DICT_START = 'd'
    LIST_START = 'l'
    DICT_LIST_END = 'e'
    DICT_KEY_VALUE_SEP = ': '
    DICT_LIST_ITEM_SEP = ', '
    INT_START = 'i'    
    
    class _TorrentStr(object):
        """ StringIO wrapper over the torrent string. 
        
            TODO:
                . Create unittests to cover this class.
                . Should this rather extend StringIO class. Explore.
        """
        
        STR_LEN_VALUE_SEP = ':'
        INT_END = 'e'    
        
        def __init__(self, torr_str):
            self.torr_str = StringIO(torr_str)
            self.curr_char = None
        
        def next_char(self):
            """
            to provide 2 ways of accessing the current parsed char - 
                1. as return value,
                2. as self.curr_char (useful in some circumstances)
            """
            self.curr_char = self.torr_str.read(1)
            return self.curr_char
        
        def step_back(self, position=-1, mode=1):
            """ Step back, by default, 1 position relative to the 
            current position.
            """
            self.torr_str.seek(position, mode) 
        
        def parse_str(self):
            """ Parse and return a string from the torrent file content. 
            Format <string length>:<string>
            
                Returns:
                    Parsed string (from the current position).
                Raises:
                    ParsingError, when expected string format is not 
                    encountered.
            
                TODO: 
                    . Explore using regex to accomplish the parsing.            
            """                                   
            str_len = self._parse_number(delimiter=self.STR_LEN_VALUE_SEP)
            
            if not str_len:
                raise ParsingError(
                    'Empty string length found while parsing at position %d'
                     % self.torr_str.pos)
            
            return self.torr_str.read(str_len)        
            
        def parse_int(self):
            """ Parse and return an integer from the torrent file content.
            Format i[0-9]+e

                Returns:
                    Parsed integer (from the current position).
                Raises:
                    ParsingError, when expected integer format is not
                    encountered.
            
                TODO: 
                    . Explore using regex to accomplish the parsing.
                    . Could re-purpose this function to parse str_length.
            """
            # just to make sure we are parsing the integer of correct format
            self.step_back()
            
            if self.next_char() != TorrentParser.INT_START:
                raise ParsingError(
                    'Error while parsing for an integer. Found %s at \
                    position %d while %s is expected.' % 
                    (self.curr_char,
                    self.torr_str.pos,
                    TorrentParser.INT_START))
                            
            return self._parse_number(delimiter=self.INT_END)
        
        def _parse_number(self, delimiter):
            """
            Parses a sequence of digits representing either an integer or
            string length and returns the number.
            """
            parsed_int = ''
            while True:
                parsed_int_char = self.next_char()
                if parsed_int_char not in string.digits:
                    if parsed_int_char != delimiter:
                        raise ParsingError(
                            'Invalid character %s found after parsing an \
                            integer (%s expected) at position %d.' % 
                            (parsed_int_char, delimiter, self.torr_str.pos))
                    else:                                          
                        break
                
                parsed_int += parsed_int_char
                
            return int(parsed_int)
       

    def __init__(self, torrent_file_path):
        """
        Reads the torrent file and sets the content as an object attribute.
        
        Args:
            torrent_file_path - String containing path to the torrent file
            to be parsed
        Returns:
            None
        Raises:
            ValueError - when passed arg is not of string type
            IOError - when the string arg passed points to a non-existent file
                    
        """
        if not isinstance(
            torrent_file_path, types.StringType) and not isinstance(
                torrent_file_path, types.UnicodeType):
            raise ValueError(
                'Path of the torrent file expected in string format.'
                )
        
        if not os.path.exists(torrent_file_path):
            raise IOError("No file found at '%s'" % torrent_file_path)
        
        self.torrent_file = open(torrent_file_path)
        self.torrent_content = self.torrent_file.read()
        self.torrent_str = self._TorrentStr(self.torrent_content)
        
        self.parsed_content = self._parse_torrent()
  
    def get_tracker_url(self):
        """ Returns the tracker URL from the parsed torrent file. """        
        return self.parsed_content.get('announce')

    
    def get_creation_date(self, time_format='iso'):
        """ Returns creation date of the torrent, if present, in ISO 
        time_format from the parsed torrent file.
            
            Args:
                time_format - determines the time_format of the time value 
                returned. 
                Valid values 'iso' or 'datetime'.
                Defaults to 'iso'.        
        """
        time_stamp = self.parsed_content.get('creation date')
        if time_stamp:
            time_stamp = datetime.utcfromtimestamp(time_stamp)
            
            if time_format == 'iso':
                return time_stamp.isoformat()
            else:
                return time_stamp
            
    
    def get_client_name(self):
        """
        Returns the name of the client that created the torrent if present,
        from the parsed torrent file.
        """
        return self.parsed_content.get('created by')

    def get_files_details(self):
        """
        Parses torrent file and returns details of the files contained
        in the torrent. Details include name, length and checksum for each
        file in the torrent.
        """
        parsed_files_info = []
        files_info = self.parsed_content.get('info')
        # 'info' should be present in all torrent files. Nevertheless.. 
        if files_info:
            multiple_files_info = files_info.get('files')
            if multiple_files_info: # multiple-file torrent
                # the name attribute was holding the directory name that each
                # of the multiple files were contained within.
                dir_name = files_info.get('name')
                for file_info in multiple_files_info:
                    # simply append the directory to the concatenated list
                    # of items under path, mostly it is a single item.
                    parsed_files_info.append(
                        (os.path.join(dir_name, os.path.sep.join(file_info.get('path'))),
                            file_info.get('length'),))
            else: # single file torrent
                parsed_files_info.append((files_info.get('name'),
                    files_info.get('length'), ))
            
        return parsed_files_info            
    
    
    def _parse_torrent(self):
        """
        Parse the torrent content in bencode format into python data format.
        
            Returns:
                A dictionary containing info parsed from torrent file.
                                
        """
        parsed_char = self.torrent_str.next_char()
        
        if not parsed_char:
             # EOF
            return
        
        # Parsing logic 
        if parsed_char == self.DICT_LIST_END: 
            return
        
        elif parsed_char == self.INT_START:
            return self.torrent_str.parse_int()
        
        elif parsed_char in string.digits: # string
            self.torrent_str.step_back()
            return self.torrent_str.parse_str()
        
        elif parsed_char == self.DICT_START:
            parsed_dict = {}
            while True:
                dict_key = self._parse_torrent()
                if not dict_key: 
                    break # End of dict                
                dict_value = self._parse_torrent() # parse value
                parsed_dict.setdefault(dict_key, dict_value)               
            
            return parsed_dict        
        
        elif parsed_char == self.LIST_START:
            parsed_list = []
            while True:
                list_item = self._parse_torrent()
                if not list_item: 
                    break # End of list  
                parsed_list.append(list_item)
            
            return parsed_list                
        
        
if __name__ == '__main__':
        
    source_dir = '/home/shad/projects/seed-sync/orig/torrentparse-master/test_data'
    files = os.listdir(u""+source_dir)

    for test_file in files:
        tp = TorrentParser(os.path.join(source_dir, test_file))
        print tp.get_tracker_url()
        print tp.get_creation_date()
        print tp.get_client_name()
        print tp.get_files_details()
        print '*' * 80




#!/usr/bin/python

"""
Task Manager
"""
from __future__ import absolute_import, print_function
import os, sys
import logging
import lockfile as locker

from seedbox import __version__
from seedbox.lockfileext import TimeoutPIDLockFile
from seedbox import logext as logmgr
from seedbox import options as opt_loader
from seedbox import pluginmanager, processmap, taskmanager, datamanager

log = logging.getLogger('main')

def _get_default_location():
    """
    get default location for running
    """
    use_resource = None
    if sys.platform.startswith('win'):
        # On windows look in ~/seedbox as well, as explorer does not let you create a 
        # folder starting with a dot
        default_path = os.path.join(os.path.expanduser('~'), 'seedbox')
    else:
        default_path = os.path.join(os.path.expanduser('~'), '.seedbox')

    # if the user folder exists; then we'll use it if we were not provided one
    if os.path.exists(default_path):
        use_resource = default_path
    else:
        # if all else fails then we just use the current working directory
        use_resource = os.getcwd()

    return use_resource

def main():
    """
    Main program
    """

    # need to create a lock to make sure multiple instances do not start at the
    # sametime because we are running as a cron.
    filelock = os.path.join(_get_default_location(), 'seedmgr.lock')
    lock = TimeoutPIDLockFile(filelock, 10)
    try:
        with lock:
            # processes all command-line inputs that control how we execute
            # logging, run mode, etc.; and we get a handle back to access the info
            configurator = opt_loader.initialize(__version__)
        
            # now retrieve processed information so we can do our setup
            core_configs = configurator.get_configs()
        
            # configure our logging
            logmgr.configure(logging.getLevelName(core_configs.loglevel.upper()),
                core_configs.logfile, core_configs.resource_path, core_configs.dev)
        
            # now that we have logging, we can process the config file
            # provided by the user; because core_configs is a reference to the
            # namespace (ala an object instance, it will automatically update,
            # so no need to acquire the core_configs again)
            configurator.load_configs()
        
            # load up all task plugins; if no plugins load then we could have a problem
            phasemap = pluginmanager.load_plugins(configurator, core_configs.plugin_paths)
        
            # now create a map between our pocess and plugins providing actions
            processmap.init(phasemap)
        
            # setup database and initialize connection; then loads up all torrents
            # into the cache. Now we have a database and potentially some torrents
            # that need processing
            datamanager.start(core_configs)
        
            # time to start processing
            taskmanager.start(core_configs)

    except locker.LockTimeout:
        # if we have managed timeout, it means there is another instance already running
        # so we will simply bow out and let the existing one still run.
        print('Unable to acquire a lock, so another process is still running. (see cron log)',
            file=sys.stderr)


if __name__ == '__main__':
    # not sure if this really works any longer due to absolute imports
    # fix it???
    main()




"""
Extends the subprocess.Popen class to provide logging threads for stdout and stderr of the child process
such that it will log stdout through logging module to provided logger.info() and log stderr to
provided logger.warn().

Because using subprocess.PIPE for stdout and stderr with any of the convience methods or with the
wait() command has potential for buffer issue, we leverage Popen directly with PIPE but then spawn
a thread with access to the corresponding logger function on the specific pipe. And then leverage
the poll method to wait for child process to complete. If the returncode is not 0, then just like
check_call() we will raise the CalledProcessError to be as consistent as possible. This is done
via the staticmethod execute(). Because we subclass Popen, you can simply create the instance and
leverage Popen methods for interaction if you prefer.

The intent here is to have the output send directly to the logger vs. being buffered until the end
with the communicate() method, or dealing with buffer conflict issues like wait(). The logger
can handle the different levels, StreamHandler or even FileHandler and the corresponding subclasses
to make it more manageable.

This was derived from several different responses on stackoverflow and blogs on this topic. After
pulling them all together and doing several rounds of testing, this was the result.
"""
import subprocess
import threading

class ProcessLogging(subprocess.Popen):
    """
    Run a command as a subprocess sending output to a logger.
    """
    def __init__(self, cmd, logger):
        """
        Create the process and start the thread for logging
        """
        self._cmd = cmd
        self._logger = logger

        # delegate to parent to spawn the rsync process
        super(ProcessLogging, self).__init__(self._cmd, shell=False,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=1, universal_newlines=True)

        # start stdout and stderr logging threads
        self.log_thread(self.stdout, self._logger.info)
        self.log_thread(self.stderr, self._logger.warn)

        self._logger.debug('Started subprocess, pid %s', self.pid)
        self._logger.debug('Command:  %s', ' '.join(self._cmd))

    def log_thread(self, pipe, log_func):
        """
        Start a thread logging output from pipe
        """
        def log_output(out, log_func):
            """
            thread function to log subprocess output
            """
            for line in iter(out.readline, b''):
                log_func(line.rstrip('\n'))

        # start thread
        thread = threading.Thread(target=log_output, args=(pipe, log_func))
        thread.setDaemon(True) # thread dies with the program
        thread.start()
        thread.join()

    def complete(self):
        """
        Handle all the processing of the subprocess
        """
        while self.poll() is None:
            # we can look to add in signal interruption handling
            # here in the future.
            pass
        else:
            # if we have a return code of anything other than 0;
            # we had some kind of failure so we should recognize the
            # error and handle accordingly.
            if self.returncode != 0:
                raise subprocess.CalledProcessError(self.returncode, self._cmd)

    @staticmethod
    def execute(cmd, logger):
        """
        provide a convenience method for creating creating the object and
        waiting for the results; very similar to check_call() but not at
        module level.
        """
        proc = ProcessLogging(cmd, logger)
        proc.complete()

# end ProcessLogging




"""
filedelete will encapsulate the deleting of files from sync directory once
synce process has completed successfully. Also managing the cache of the file
and corresponding state.
"""
from __future__ import absolute_import
import logging
import os

from seedbox import helpers
from seedbox.pluginmanager import register_plugin, phase

__version__ = '0.1'

log = logging.getLogger(__name__)

class DeleteFile(object):
    """
    Encapsulates the delete command
    """

    @phase(name='complete')
    def execute(self, torrents, configs):
        """
        perform the delete of a file
        """
        processed_torrents = []

        log.trace('starting delete file plugin.')
        for torrent in torrents:

            try:
                log.debug('deleting media files for torrent %s', torrent)
                media_files = helpers.get_media_files(torrent, synced=True)
    
                # now loop through the files we got back, if none then no files
                # were in need of deleting
                for media_file in media_files:
                    # if for some reason it doesn't exist then no worries just skip
                    # over it and continue; otherwise delete it
                    if os.path.exists(os.path.join(configs.sync_path, media_file.filename)):
                        log.trace('delete file: %s', media_file.filename)
                        os.remove(os.path.join(configs.sync_path, media_file.filename))
                
                # after we are done processing the torrent added it the list of
                # processed torrents
                processed_torrents.append(torrent)

            except Exception as err:
                log.info('%s was unable to process %s due to [%s]', DeleteFile.__name__, torrent, err)
                # TODO: need to refine this further so we know what errors really happened
                helpers.set_torrent_failed(torrent, err)

        log.trace('ending delete file plugin')
        return processed_torrents

register_plugin(DeleteFile)




"""
filecopy will encapsulate the copying of files from completed directory to
sync directory when file is not in a compressed format. Also managing the
cache of the file and corresponding state.
"""
from __future__ import absolute_import
import logging
import os
import shutil

from seedbox import helpers
from seedbox.pluginmanager import register_plugin, phase

__version__ = '0.1'

log = logging.getLogger(__name__)

class CopyFile(object):
    """
    Encapsulates the copy command
    """
    
    # our list of supported extensions, no need to go copying all files
    supported_exts = ['.mp4', '.avi']

    @phase(name='prepare')
    def execute(self, torrents, configs):
        """
        perform the copy of a file
        """
        processed_torrents = []

        log.trace('starting copy file plugin')
        for torrent in torrents:

            try:
                log.debug('copying media files for torrent %s', torrent)
                media_files = helpers.get_media_files(torrent, file_exts=CopyFile.supported_exts)
    
                # now loop through the files we got back, if none then no files
                # were in need of copying
                for media_file in media_files:
                    log.trace('copying file: %s', media_file.filename)
                    shutil.copy2(os.path.join(media_file.file_path, media_file.filename), configs.sync_path)

                    # we are copying the files not the paths; if the filename includes
                    # path information we will need to strip it for future consumption when we
                    # create the new file entry. The helper method expects a list but in our
                    # case it is just a single file so wrap it up!
                    helpers.add_mediafiles_to_torrent(torrent, configs.sync_path,
                        [os.path.basename(media_file.filename)])
                    # need to show we completed the copy process; so show it as synced
                    # which is technically true since we synced it to the sync path.
                    helpers.synced_media_file(media_file)

                
                helpers.set_media_files_path(configs.sync_path, media_files)

                # after we are done processing the torrent added it the list of
                # processed torrents
                processed_torrents.append(torrent)

            except Exception as err:
                log.info('%s was unable to process %s due to [%s]', CopyFile.__name__, torrent, err)
                # TODO: need to refine this further so we know what errors really happened
                helpers.set_torrent_failed(torrent, err)

        log.trace('ending copy file plugin')
        return processed_torrents

register_plugin(CopyFile)




"""
filesync was inspired by a posting for doing something similar
@ http://filesharingtalk.com/threads/436669-One-way-syncing-downloads-from-seedbox-to-home

The filesync module encapsulates the rsync command for doing a sync between
source and destination. Also tracking what has been sync'd given the nature of
seedboxes where things stay for extended periods but the destination files
are usually moved into media libraries. 

Provided is the ability to set a subset of the rsync options based on the
most common needs. Flags will be either True/False (None not supported)

    dryrun (rsync: dry-run): default[False]
    verbose: default[False]
    progress: default[False]
    perms: default[True]
    delayupdates (rsync: delay-updates): default[True]
    recursive: default[True]
    chmod: default[ugo+rwx]
    identity (rsync: rsh=ssh) default[]
    port (rsync: rsh=ssh) default[] rsyc default[22]

* assumes settings is a key:value (dict) of the possible configurations.     
"""
from __future__ import absolute_import
import logging
import os

from seedbox import helpers
from seedbox.pluginmanager import register_plugin, phase
from seedbox.subprocessext import ProcessLogging

__version__ = '0.1'

log = logging.getLogger(__name__)

class SyncFile(object):
    """Encapsulates the rsync command with the additional feature of having a
    database cache to track which files were already successfully processed
    so they are not re-processed on future executions when the destination
    copy of the file has been moved or deleted.

    Provided is a subset of the inputs the rsync command actually supports
    but these are the most common ones required.

    dryrun (rsync: dry-run)
    verbose
    progress
    perms
    delayupdates (rsync: delay-updates)
    recursive
    chmod
    logfile (rsync: log-file)
    identity (rsync: rsh=ssh)
    port (rsync: rsh=ssh)
    """
    def __init__(self):
        """
        Establish instance variables that can be updated via configure() method
        at a later point
        """
        self._cmd = None
        self.attrs = {
            'dryrun': False,
            'verbose': False,
            'progress': False,
            'perms': True,
            'delayupdates': True,
            'recursive': True,
            'chmod': None,
            'identity': None,
            'port': None,
            'sync_path': None,
            'remote_path': None,
        }

    @phase(name='activate')
    def execute(self, torrents, configs):
        """
        perform the rsync of a file
        """
        processed_torrents = []

        log.trace('starting sync file plugin')
        self._configure(configs)

        for torrent in torrents:

            try:
                log.debug('syncing media files for torrent %s', torrent)
                media_files = helpers.get_media_files(torrent, synced=False)
    
                # now loop through the files we got back, if none then no files
                # were in need of syncing
                for media_file in media_files:
                    log.trace('syncing file: %s', media_file.filename)
                    self._sync(media_file.filename)
                    helpers.synced_media_file(media_file)
                
                # after we are done processing the torrent added it the list of
                # processed torrents
                processed_torrents.append(torrent)

            except Exception as err:
                log.info('%s was unable to process %s due to [%s]', SyncFile.__name__, torrent, err)
                # TODO: need to refine this further so we know what errors really happened
                helpers.set_torrent_failed(torrent, err)

        log.trace('ending sync file plugin')
        return processed_torrents


    def _set_cmd(self):
        """Generate the rsync command using the instance variables as
        configured by user input or using the defaults we set. 

        Args:
            N/A

        Returns:
            N/A
        """
        self._cmd = ['rsync']

        if self.attrs.get('dryrun'):
            self._cmd.append('--dry-run')
        if self.attrs.get('verbose'):
            self._cmd.append('--verbose')
        if self.attrs.get('progress'):
            self._cmd.append('--progress')
        if self.attrs.get('perms'):
            self._cmd.append('--perms')
        if self.attrs.get('delayupdates'):
            self._cmd.append('--delay-updates')
        if self.attrs.get('recursive'):
            self._cmd.append('--recursive')
        if self.attrs.get('chmod'):
            self._cmd.append('--chmod=%s' % self.attrs.get('chmod'))

        if self.attrs.get('identity') or self.attrs.get('port'):
            ssh_cmd = '--rsh=ssh'
            if self.attrs.get('port'):
                ssh_cmd = ssh_cmd + ' -p ' + self.attrs.get('port')
            if self.attrs.get('identity'):
                ssh_cmd = ssh_cmd + ' -i ' + self.attrs.get('identity')

            self._cmd.append(ssh_cmd)

        log.debug('formatted cmd: [%s]', self._cmd)


    def _configure(self, configs):
        """Settings all the instance variables based on inputs provided
        by the caller. Given there is such a large number and it could
        grow over time, we support key:value pair (dict) input style.
        All flags require an input of True/False, None not allowed.

        Args:
            kwargs: key:value pair inputs for the support instance variables

        Return:
            N/A
        """
        
        # it doesn't mean any values will be actually change just most likely
        for attr_key in self.attrs.keys():
            log.trace('checking key: %s', attr_key)
            if hasattr(configs, attr_key):
                log.trace('key found in config: %s', attr_key)
                config_val = getattr(configs, attr_key)
                log.trace('value from config: %s', config_val)
                if config_val is not None:
                    log.trace('value from config is not None; setting as our attr')
                    self.attrs[attr_key] = config_val

        # now update cmd based on updated configurations
        self._set_cmd()


    def _sync(self, filename):
        """Performs actual rsync between 2 locations.

        Args:
            source_dir: source directory containing files to sync
            destination_dir: destination directory where files will be synced

        Returns:
            N/A
        """
        # Sync the file.
        log.info('syncing new file %s', filename)
        log.debug('base sync command [%s]', self._cmd)

        # since we need to modify the command to add iteration specific
        # elements we will capture the base command into a local variable
        # must not forget to clone, stupid pass by reference!!!
        cmd = self._cmd[:]

        cmd.append(os.path.join(self.attrs.get('sync_path'), filename))
        cmd.append(self.attrs.get('remote_path') + os.sep)
        log.debug('file specific command [%s]', cmd)

        ProcessLogging.execute(cmd, log)
        log.info('file synced %s', filename)


register_plugin(SyncFile)




"""
filedecompress ecapsulates the decompressing (unrar) command of files
in the completed directory to the sync directory. Also managing the
cache of the file and corresponding state.
"""
from __future__ import absolute_import
import logging
import os
import rarfile

from seedbox import helpers, tools
from seedbox.pluginmanager import register_plugin, phase

__version__ = '0.1'

log = logging.getLogger(__name__)

# make sure the path for unrar is set properly; if we can't find it
# then we'll just have to let it error out and the user will need to figure
# it out on their system.
unrar_exec = tools.get_exec_path('unrar')
if unrar_exec is not None:
    rarfile.UNRAR_TOOL = unrar_exec

class UnrarFile(object):
    """
    Encapsulates the unrar command
    """

    @phase(name='prepare')
    def execute(self, torrents, configs):
        """
        perform the unrar of a file
        """
        processed_torrents = []

        log.trace('starting unrar file plugin.')
        for torrent in torrents:

            try:
                log.debug('unrar media files for torrent %s', torrent)
                media_files = helpers.get_media_files(torrent, compressed=True)
    
                # now loop through the files we got back, if none then no files
                # were in need of unraring
                for media_file in media_files:
                    log.trace('unrar file: %s', media_file.filename)
                    with rarfile.RarFile(os.path.join(media_file.file_path, media_file.filename)) as compressed_file:
                        archived_files = compressed_file.namelist()
                        compressed_file.extractall(path=configs.sync_path)
                        helpers.add_mediafiles_to_torrent(torrent, configs.sync_path, archived_files)
                        # need to show we completed the unrar process; so show it as synced
                        # which is technically true since we synced it to the sync path.
                        helpers.synced_media_file(media_file)

                
                # after we are done processing the torrent added it the list of
                # processed torrents
                processed_torrents.append(torrent)
    
            except Exception as err:
                log.info('%s was unable to process %s due to [%s]', UnrarFile.__name__, torrent, err)
                # TODO: need to refine this further so we know what errors really happened
                helpers.set_torrent_failed(torrent, err)

        log.trace('ending unrar file plugin')
        return processed_torrents


register_plugin(UnrarFile)




#!/usr/bin/env python

import os, re, sys

try:
    # make sure setuptools are available
    from ez_setup import use_setuptools
    use_setuptools()
except ImportError:
    # just ignore it
    pass

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

PKG_DIR='src'

# handle dynamically getting the version
# for our application
VERFILE = os.path.join(PKG_DIR, 'seedbox', '_version.py')
verstr = 'unknown'
try:
    verstrline = open(VERFILE, 'rt').read()
except EnvironmentError as err:
    pass    # file is missing
else:
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise RuntimeError('if {0} exists, it is required to be well-formed'.format(VERFILE))

# generate a list of required components
required = []
required.append('SQLObject >= 1.3.2')
required.append('xworkflows')
required.append('rarfile >= 2.6')
required.append('shutilwhich')
required.append('lockfile')

data_files = ['LICENSE', 'README', 'README.md']

setup(
      name='SeedboxManager',
      version=verstr,
      description='Provides framework for performing tasks on a seedbox',
      author='shad7',
      author_email='kenny.shad7@gmail.com',
      url='https://github.com/shad7/seedbox',
      license='MIT',
      packages=find_packages(PKG_DIR),
      package_dir={'': PKG_DIR},
      data_files=data_files,
      install_requires=required,
      entry_points={ 'console_scripts': ['seedmgr = seedbox.manager:main'] },
      zip_safe=False,
      tests_require=['pytest'],
      cmdclass = {'test': PyTest},
)




"""
Handles all interaction with data model and related persistence.
Basically it provides an interface for all actions you can do to the
objects within our model. Otherwise we could end up with challenges
related to data persistence or even having to understand how things are
persisted outside of the model area.
"""
from __future__ import absolute_import
import datetime as datemod
import logging
import os
import time
from seedbox import tools
from seedbox.model import schema
from seedbox.torrentparser import TorrentParser, MalformedTorrentError, ParsingError

log = logging.getLogger(__name__)

def add_torrent(name):
    """
    Create the torrent and return it
    """
    torrent = schema.Torrent(name=name)

    return torrent

def update_state(torrents, state):
    """
    update the list of torrents state attribute to the new state (bulk style)
    """
    # there has to be a faster method for doing this in bulk w/o
    # enabling lazy updates for the entire object...blah!!!
    # TODO: investigate further
    for torrent in torrents:
        # we'll progress the state of the torrent only if processing did not
        # result in an error. Such that we need to go through retry logic
        if not torrent.failed:
            torrent.state = state

def set_failed(torrent, exception):
    """
    Update torrent to a failed/error state
    """
    torrent.failed = True
    torrent.error_msg = exception

def reset_failed(torrent):
    """
    reset a torrent such that it is eligible to reprocess again
    """
    torrent.failed = False
    torrent.error_msg = None

def fetch_torrent_by_name(name):
    """
    Retrieve torrent by name
    """
    search = schema.Torrent.selectBy(name=name)
    # because name is unique, it will always be Zero or One entry
    # so we can use the getOne() feature. By passing in None we avoid
    # get back an exception, and therefore we can check for no
    # torrent and simply create it.
    torrent = search.getOne(None)

    return torrent


def fetch_or_create_torrent(name):
    """
    Attempt to fetch the torrent, if not found, then create it
    """
    torrent = fetch_torrent_by_name(name)

    if not torrent:
        log.debug('Torrent [%s] not found; Creating torrent instance...', name)
        torrent = add_torrent(name)

    return torrent


def add_file_to_torrent(torrent, media):
    """
    Add a file to the torrent
    """
    schema.MediaFile(filename=media.get('filename'), file_ext=media.get('file_ext'),
        file_path=media.get('file_path'),
        size=media.get('size'), compressed=media.get('compressed'),
        synced=media.get('synced'), missing=media.get('missing'),
        skipped=media.get('skipped'), torrent=torrent)


def add_files_to_torrent(torrent, file_list):
    """
    Add files to the torrent
    """
    for media in file_list:
        add_file_to_torrent(torrent, media)


def get_files_by_torrent(torrent):
    """
    retrieves the list of files for a given torrent
    """
    return list(torrent.media_files)


def get_files_by_torrentname(name):
    """
    simple convience method for get files with only a torrent name
    """
    return get_files_by_torrent(fetch_torrent_by_name(name))


def get_torrents_by_state(state, failed=False):
    """
    For a given state, retrieve a list of torrents;  by default ignore those
    that are still need to go through retry, but if requested get those
    that need to be retried for a given state.
    """
    search = schema.Torrent.selectBy(state=state, failed=failed)

    return list(search)

def get_torrents_to_retry():
    """
    To all the torrents that have not been aborted/cancelled that are in
    and error state.
    """
    search = schema.Torrent.selectBy(failed=True)

    return list(search)

def delete_torrents(torrents):
    """
    Delete a list of torrents
    """
    for torrent in torrents:
        schema.Torrent.delete(id=torrent.id)

def delete_media_files(media_files):
    """
    Delete a list of media files
    """
    for media_file in media_files:
        schema.MediaFile.delete(id=media_file.id)

def get_eligible_for_purging():
    """
    get list of torrents for which the mediafiles are eligible for purging
    """
    where_clause = []
    where_clause.append('torrent.invalid = 0')
    where_clause.append('AND torrent.purged = 0')
    where_clause.append('AND (torrent.state = {0}'.format(schema.Torrent.sqlrepr('done')))
    where_clause.append('OR torrent.state = {0})'.format(schema.Torrent.sqlrepr('cancelled')))
    str_where_clause = ' '.join(where_clause)

    log.debug('WHERE: [%s]', str_where_clause)
    search = schema.Torrent.select(str_where_clause, distinct=True)

    return list(search)

def get_eligible_for_removal():
    """
    get list of torrents for which could be eligible for actual removal
    """
    where_clause = []
    where_clause.append('torrent.purged = 1')
    where_clause.append('AND (torrent.state = {0}'.format(schema.Torrent.sqlrepr('done')))
    where_clause.append('OR torrent.state = {0})'.format(schema.Torrent.sqlrepr('cancelled')))
    str_where_clause = ' '.join(where_clause)

    log.debug('WHERE: [%s]', str_where_clause)
    search = schema.Torrent.select(str_where_clause, distinct=True)

    return list(search)

###
# Helper private implementations; see seedbox.helpers for public interface
# to these private interfaces. All validation, default values, and other
# checks are handled on the public side so we can keep our interface and
# implementation as clean and simple as possible. Also then if someone
# wants to see what is going on, they can see checks there vs. digging around
# in our private implementation.
###

def get_media_files(torrent, compressed, synced, missing, skipped):
    """
    implementation for helpers; the public interface is provided
    by helpers so we don't provide any default values on this interface
    """
    search = schema.MediaFile.selectBy(torrent=torrent.id, compressed=compressed,
        synced=synced, missing=missing, skipped=skipped)

    return list(search)

def get_processed_media_files(torrent):
    """
    Get all media files that fall in the category of processed
    synced = True; missing = True; or skipped = True
    """
    # if the torrent is been listed as invalid or has been purged
    # already then no need to query for more data.
    if torrent.invalid or torrent.purged:
        log.debug('no media files to retrieve as torrent is invalid or already purged')
        return []

    where_clause = []
    where_clause.append('media_file.torrent_id = %d' % torrent.id)
    where_clause.append('AND (media_file.synced = 1')
    where_clause.append('OR media_file.missing = 1')
    where_clause.append('OR media_file.skipped = 1)')
    str_where_clause = ' '.join(where_clause)

    log.debug('WHERE: [%s]', str_where_clause)
    search = schema.MediaFile.select(str_where_clause, distinct=True)

    return list(search)

def purge_media(torrent):
    """
    Delete the corresponding media files from the table based on
    the specified torrent (means all processing is done, so no need to keep
    track of all the files originally associated with torrent.
    """
    schema.MediaFile.deleteBy(torrent=torrent.id)
    torrent.purged = True

def backup_db(resource_path):
    """
    perform backup of the database, typical right before a purge action
    """
    schema.backup(resource_path)

###
# Functions for managing the application state data entries
###
def _fetch(name):
    """
    Checks to see if the name already exists in db
    """
    search = schema.AppState.selectBy(name=name)
    # because name is unique, it will always be Zero or One entry
    # so we can use the getOne() feature. By passing in None we avoid
    # get back an exception, and therefore we can check for no
    # results and simply create it if needed
    entry = search.getOne(None)

    return entry

def _create(name):
    """
    Creates the new entry in the app state
    """
    return schema.AppState(name=name)

def _fetch_or_create(name):
    """
    convience function that handles retrieving existing entry or creating
    a new one for us.
    """
    entry = _fetch(name)

    if not entry:
        entry = _create(name)

    return entry

def set(name, value):
    """
    set name to value where value is stored as string
    """
    entry = _fetch_or_create(name)
    entry.val_str = value

def set_int(name, value):
    """
    set name to value where value is stored as int
    """
    entry = _fetch_or_create(name)
    entry.val_int = tools.to_int(value)

def set_list(name, values):
    """
    set name to value where value is stored as delimited list
    """
    entry = _fetch_or_create(name)
    entry.val_list = tools.list_to_str(values)

def set_flag(name, value):
    """
    set name to value where value is stored as boolean
    """
    entry = _fetch_or_create(name)
    entry.val_flag = tools.to_bool(value)

def set_date(name, value):
    """
    set name to value where value is stored as date; if date is None
    we will let it default to current date.
    """
    entry = _fetch_or_create(name)
    if value and isinstance(value, datemod.datetime):
        entry.val_date = value
    else:
        entry.val_date = datemod.datetime.utcnow()

def get(name, default=None):
    """
    retrieve named value where value is stored as string
    """
    entry = _fetch(name)

    if not entry:
        return default
    else:
        return entry.val_str

def get_int(name, default=None):
    """
    retrieve named value where value is stored as int
    """
    entry = _fetch(name)

    if not entry:
        return default
    else:
        return entry.val_int

def get_list(name, default=None):
    """
    retrieve named value where value is stored as delimited list of values
    """
    entry = _fetch(name)

    if not entry:
        return default
    else:
        return tools.to_list(entry.val_list)

def get_flag(name, default=None):
    """
    retrieve named value where value is stored as boolean
    """
    entry = _fetch(name)

    if not entry:
        return default
    else:
        return entry.val_flag

def get_date(name, default=None):
    """
    retrieve named value where value is stored as date
    """
    entry = _fetch(name)

    if not entry:
        return default
    else:
        return entry.val_date

###
# This is where we handle the loading and/or initializing of the database and all the
# torrents and associated media files.
###
def start(configs):
    """
    Make sure the database is configured and connection established
    """
    schema.init(configs.resource_path, configs.reset)

    # step to clean up database
    perform_db_cleanup(configs.resource_path, configs.torrent_path, configs.reset)

    # as part of running in retry mode; we don't want to load anything
    # new so we will skip the load aspect and only handle the database
    # aspect of starting up
    if not configs.retry:
        load_torrents(configs.torrent_path, configs.media_paths, configs.incomplete_path)

def perform_db_cleanup(resource_path, torrent_location, reset_flag=False):
    """
    Determine if it is time to clean up the database, if it is then backup database and 
    start purging the necessary data.
    """
    default_date = datemod.datetime.min
    state_key = 'last_purge_date'
    one_week = datemod.timedelta(weeks=1)
    
    log.trace('starting perform_db_cleanup...')

    if reset_flag:
        log.debug('reset performed....need to reset app state. Setting last purge date to now.')
        set_date(state_key, None)
        return

    last_purge_date = get_date(state_key, default_date)
    log.debug('last_purge_date: %s', last_purge_date)

    if last_purge_date == default_date:
        log.info('First running...setting last purge date to now')
        # never been purged so need to set an initial date (today)
        set_date(state_key, None)
        return

    log.debug('current_date - last_purge_date vs. 1 week: (%s) - (%s) >= (%s) = (%s)', 
        datemod.datetime.utcnow(), last_purge_date, one_week, datemod.datetime.utcnow() - last_purge_date)
    if (datemod.datetime.utcnow() - last_purge_date) >= one_week:
        log.trace('last purge was more than 1 week ago....ready for some clean up.')
        # ready to start purge process....
        torrents = get_eligible_for_purging()
        if torrents:
            log.trace('found torrents eligible for purging: %s', len(torrents))
            # perform database backup
            backup_db(resource_path)

            log.debug('purging media associated with torrents....')
            for torrent in torrents:
                purge_media(torrent)

            # done deleting mediafiles; now delete any torrents that were found to be
            # missing from filesystem thereby no longer in need of caching.
            torrents = get_eligible_for_removal()
            torrents_to_delete = []
            for torrent in torrents:
                if not os.path.exists(os.path.join(torrent_location, torrent.name)):
                    # actual torrent file no longer exists so we can safely delete
                    # the torrent from cache
                    torrents_to_delete.append(torrent)

            log.debug('found torrents eligible for removal: %s', len(torrents_to_delete))
            delete_torrents(torrents_to_delete)

            # now reset the last purge date to now.
            set_date(state_key, None)

        else:
            # no torrents eligible for purging
            log.debug('no torrents found eligible for purging...')
    
    log.trace('perform_db_cleanup completed')    

def load_torrents(torrent_location, media_locations, inprogress_location):
    """
    Find all the torrents in the specified directory, verify it is a valid
    torrent file (via parsing) and capture the relevant details. Next create
    a record in the cache for each torrent.
    """
    torrent_files = os.listdir(u""+torrent_location)

    log.trace('torrent files found [%d] processing', len(torrent_files))
    start_time = time.time()
    total_media_files = 0

    for torrent_file in torrent_files:
        ext = os.path.splitext(torrent_file)[1]
        # some torrents don't always end with .torrent,
        # so let the parser reject it
        if  not ext or ext == '.torrent':
            # get the entry in the cache or creates it if it doesn't exist
            torrent = fetch_or_create_torrent(torrent_file)

            if _is_parsing_required(torrent):

                try:
                    parser = TorrentParser(os.path.join(torrent_location, torrent.name))
                    media_items = parser.get_files_details()
                    log.trace('Total files in torrent %d', len(media_items))
                    total_media_files += len(media_items)

                    # determine if any of the files are still inprogress of being
                    # downloaded; if so then go to next torrent. Because no files
                    # were added to the cache for the torrent we will once again
                    # parse and attempt to process it.
                    if _is_torrent_downloading(inprogress_location, media_items):
                        log.trace('torrent still downloading, next...')
                        continue

                    media_files = _parse_torrent(media_locations, media_items)
                    add_files_to_torrent(torrent, media_files)
                except MalformedTorrentError as mte:
                    torrent.invalid = 1
                    # probably not wise to do this, but better than trying to
                    # include multiple conditions in queries when this would
                    # immeidiately filter out the truly invalid torrents when
                    # our workflow process is executing torrents based on current
                    # state of the torrent (state = workflow state)
                    torrent.state = 'cancelled'
                    log.error(
                        'Malformed Torrent: [%s] [%s]', torrent_file, mte)
                except ParsingError as ape:
                    torrent.invalid = 1
                    # probably not wise to do this, but better than trying to
                    # include multiple conditions in queries when this would
                    # immeidiately filter out the truly invalid torrents when
                    # our workflow process is executing torrents based on current
                    # state of the torrent (state = workflow state)
                    torrent.state = 'cancelled'
                    log.error('Torrent Parsing Error: [%s] [%s]', 
                        torrent_file, ape)

        else:
            log.warn('Skipping unknown file: [%s]', torrent_file)

    took = time.time() - start_time
    log.debug('torrents and media files [%d] took %.2f seconds to load', total_media_files, took)


def _is_parsing_required(torrent):
    """
    Determines if parsing is required. Checks the following attributes 
    to determine if we should skip parsing or not:
    - if torrent.invalid is True, skip parsing
    - if torrent.purged is True, skip parsing
    - if len(torrent.media_files) > 0, skip parsing
    """
    parse = True

    # if we have parse the file previous and marked it invalid, then no
    # need to bother doing it again
    if torrent.invalid is True:
        parse = False

    # if we have already purged the torrent media files, then no need
    # to perform any parsing again
    if torrent.purged is True:
        parse = False

    # we have already parsed the file previously, which is why there are
    # files already associated to the torrent, otherwise it would have
    # been zero
    if torrent.media_files.count() > 0:
        parse = False

    return parse

def _is_torrent_downloading(inprogress_location, media_items):
    """
    Verify if atleast one item still located in the inprogress/downloading location
    
    args:
        inprogress_location: the directory/path to where files are stored while downloading happens
        media_items: files found inside a torrent
    returns:
        true: if any file is still within inprogress_location
        false: if no file is within inprogress_location
    """

    found = False
    for (filename, filesize) in media_items:

        # if the file is found then break out of the loop and
        # return found; else we will return default not found
        if os.path.exists(os.path.join(inprogress_location, filename)):
            found = True
            break

    return found


def _parse_torrent(media_locations, media_items):
    """
    Handles interacting with torrent parser and getting required details
    from the parser.

    args:
        torrent: includes access to torrent and its location
    """

    file_list = []
    for (filename, filesize) in media_items:
        if filename:

            details = {}
            details['filename'] = filename
            details['size'] = filesize

            in_ext = os.path.splitext(filename)[1]
            details['file_ext'] = in_ext

            # default to missing; no need to check all the paths for a file
            # if we don't really care about the file to start with, ie a file
            # we plan to skip
            details['file_path'] = None

            # we will assume each file is not synced
            # desired (not skipped), and accessible
            details['compressed'] = 0
            details['synced'] = 0
            details['skipped'] = 0
            details['missing'] = 0

            # if ends with rar, then it is a compressed file;
            if in_ext in ('.rar'):
                details['compressed'] = 1

            # if the file is a video type, but less than 75Mb then
            # the file is just a sample video and as such we will
            # skip the file
            elif in_ext in ('.mp4', '.avi'):
                if filesize < 75000000:
                    details['skipped'] = 1
                    log.debug(
                        'Based on size, this is a sample file [%s].\
                         Skipping..', filename)

            # if the file has any other extension (rNN, etc.) we will
            # simply skip the file
            else:
                details['skipped'] = 1
                log.debug(
                    'Unsupported filetype (%s); skipping file', in_ext)

            # now that we know if the file should be skipped or not from above
            # we will determine if the file is truly available based on the
            # full path and filename. If file is not found, then we will set
            # the file to missing
            if not details.get('skipped'):
                details['file_path'] = _get_file_path(media_locations, filename)
                if not details['file_path']:
                    details['missing'] = 1
                    log.info('Media file [%s] not found', filename)

            # now we can add the file to the torrent
            file_list.append(details)
            log.trace('Added file to torrent with details: [%s]', details)

        else:
            raise MalformedTorrentError('No indexed files found in torrent.')

    return file_list

def _get_file_path(media_locations, filename):
    """
    A list of locations/paths/directories where the media file could exist.

    return:
        location if found
        None if not found
    """

    found_location = None
    for location in media_locations:

        # if the file is found then break out of the loop and
        # return found; else we will return default not found
        if os.path.exists(os.path.join(location, filename)):
            found_location = location
            break

    return found_location




"""
These are helper functions that individual tasks can leverage to perform some
pre-defined activities without having to understand how we do caching, updates,
or even what the process flow is. By hiding all of these details we can keep
internals private.
"""
from __future__ import absolute_import
from itertools import ifilter
import logging
import os
from seedbox import datamanager

log = logging.getLogger(__name__)

def set_torrent_failed(torrent, error):
    """
    set the torrent as failed and corresponding error

    args:
        torrent: torrent entry provided as input to all plugins
        error: the actual exception that happened        
    """
    datamanager.set_failed(torrent, error)

def get_media_files(torrent, file_exts=None, compressed=False, synced=False,
    missing=False, skipped=False):
    """
    Retrieve mediafiles associated with the specified torrent entry.

    args:
        torrent: torrent entry provided as input to all plugins

        (optional filters inputs)
        file_ext: list of file extensions to include in the results list
        compressed: flag to include or not include compressed files
        synced: flag to include or not include files synced
        missing: flag to include or not include files there were not able to be located during parsing
        skipped: flag to include or not include files that were ignored during parsing (extra type files)
    return:
        list of mediafiles or empty list if no files found matching criteria
    exceptions:
        ValueError: when incorrect inputs provided (eg. missing torrent or empty torrent)
    """

    log.trace('looking for media files for torrent %s', torrent)

    if not torrent or not torrent.id:
        raise ValueError('missing input: torrent is a required input')

    if compressed is None or synced is None or missing is None or skipped is None:
        raise ValueError('invalid input: flag inputs must be boolean (True/False)')

    media_files = datamanager.get_media_files(torrent, compressed, synced, missing, skipped)
    if not media_files:
        log.debug('no mediafiles found for specified torrent')
        # no results found
        return []

    if file_exts:
        log.trace('filtering list of results using list of exts %s', file_exts)
        ext_check = lambda media: media.file_ext in file_exts
        # now filter out any files that don't match the supplied file_ext list
        # also need to perform list otherwise it is an ifilter iterable
        return list(ifilter(ext_check, media_files))

    log.debug('found %d media files', len(media_files))
    return media_files

def synced_media_file(media_file):
    """
    perform sync of media file into the cache

    args:
        media_file: a media file entry retrieved when calling get_media_files
    returns:
        N/A
    exceptions:
        ValueError: when incorrect inputs provided (eg. missing media_file or empty media_file)
    """
    if not media_file or not media_file.id:
        raise ValueError('missing input: torrent is a required input')

    log.trace('syncing media file to cache %s', media_file)
    media_file.synced = True
    log.trace('synced to cache %s', media_file)

def set_media_files_path(file_location, media_files):
    """
    Update the file path for each of the files in the list.

    args:
        file_location: the path/directory to where the files are stored on disk
        media_files: the cached files that have an updated location
    returns:
        N/A
    exceptions:
        OSError: when a file in the list of filenames doesn't exist
        TypeError: when added files is not a list
        ValueError: when incorrect inputs provided; including missing inputs
    """

    # if the file_location wasn't provided/empty, or doesn't actually exist, and is not a directory
    # then we will raise an exception since we can't proceed w/o this
    if not file_location or not os.path.exists(file_location) or not os.path.isdir(file_location):
        raise ValueError('invalid input file_location provided')

    # if provided and not a list type; then we have an issue
    if media_files and not isinstance(media_files, list):
        raise TypeError('invalid input type for media_files provided: expected a list')

    log.trace('setting location [%s] for %d media files', file_location, len(media_files))
    for media_file in media_files:
        media_file.file_path = file_location

    log.trace('location set')


def add_mediafiles_to_torrent(torrent, file_location, added_files):
    """
    Add mediafiles to cache associated to torrent after decompressing an archive.

    args:
        torrent: torrent entry provided as input to all plugins
        file_location: directory/path where the file was decompressed to
        added_files: list of filenames that need to be processed and cached
    returns:
        N/A
    exceptions:
        OSError: when a file in the list of filenames doesn't exist
        TypeError: when added files is not a list
        ValueError: when incorrect inputs provided; including missing inputs
    """

    if not torrent or not torrent.id:
        raise ValueError('missing input: torrent is a required input')

    # if the file_location wasn't provided/empty, or doesn't actually exist, and is not a directory
    # then we will raise an exception since we can't proceed w/o this
    if not file_location or not os.path.exists(file_location) or not os.path.isdir(file_location):
        raise ValueError('invalid input file_location provided')

    # if provided and not a list type; then we have an issue
    if added_files and not isinstance(added_files, list):
        raise TypeError('invalid input type for added_files provided: expected a list')

    media_files = []

    log.trace('start processing files to be added to torrent %s', torrent)
    for added_file in added_files:

        media = {}
        # as part of getting size it checks for the existence of file
        # if it doesn't exist it will raise an OSError()
        media['size'] = os.path.getsize(os.path.join(file_location, added_file))
        media['filename'] = added_file
        media['file_ext'] = os.path.splitext(added_file)[1]
        media['file_path'] = file_location
        media['compressed'] = 0
        media['synced'] = 0
        media['skipped'] = 0
        media['missing'] = 0

        media_files.append(media)

    log.trace('total media files to add %d', len(media_files))
    if media_files:
        log.trace('adding files to torrent')
        datamanager.add_files_to_torrent(torrent, media_files)
        log.trace('files added')

